/**h* DesfireAPI/Legacy
 *
 * NAME
 *   DesfireAPI :: Specific entry points for Legacy re-entrant and non-reentrant DLL
 *
 * COPYRIGHT
 *   (c) 2009 SpringCard - www.springcard.com
 *
 * SEE ALSO
 *   LegacyEX
 *
 **/

#include "sprox_desfire_i.h"

#ifndef SPROX_DESFIRE_PCSC

/**f* DesfireAPI/[Legacy]SelectCid
 *
 * NAME
 *   [Legacy]SelectCid
 *
 * DESCRIPTION
 *   Selects the logical number of the addressed DesFire card
 *   CID is used to distinguish several Desfire cards simultaneously selected by a single reader
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_SelectCid (BYTE cid);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_SelectCid(SPROX_INSTANCE rInst, BYTE cid);
 *
 *   [[pcsc_desfire.dll]]
 *   Not applicable.
 * 
 * INPUTS
 *   BYTE cid        : logical number of the addressed DesFire card
 *
 * RETURNS
 *   DF_OPERATION_OK : CID selected
 *
 **/
SPROX_API_FUNC(Desfire_SelectCid) (SPROX_PARAM  BYTE cid)
{
  SPROX_DESFIRE_GET_CTX();
  ctx->tcl_cid = cid;
  return DF_OPERATION_OK;
}


/**f* DesfireAPI/[Legacy]IsoWrapping
 *
 * NAME
 *   [Legacy]IsoWrapping
 *
 * DESCRIPTION
 *   Select the wrapping mode of Desfire legacy commands into ISO 7816-4 APDUs.
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_IsoWrapping (BYTE mode);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_IsoWrapping(SPROX_INSTANCE rInst, BYTE mode);
 *
 *   [[pcsc_desfire.dll]]
 *   Not applicable. See Desfire_IsoWrapping[PC/SC].
 *
 * INPUTS
 *   BYTE mode             : ISO wrapping mode (DF_ISO_WRAPPING_CARD or DF_ISO_WRAPPING_READER)
 *
 * RETURNS
 *   DF_OPERATION_OK       : ISO wrapping mode set
 *   DFCARD_LIB_CALL_ERROR : ISO wrapping mode not set, mode not available
 *
 * NOTES
 *   DF_ISO_WRAPPING_CARD uses the card's wrapping method (see chapter 3.13 
 *   of datasheet of mifare DesFire MF3ICD40 for more information).
 *
 *   DF_ISO_WRAPPING_OFF disables the wrapping (native legacy command are used).
 *   This is the default mode.
 *
 **/

SPROX_API_FUNC(Desfire_IsoWrapping) (SPROX_PARAM  BYTE mode)
{
  SPROX_DESFIRE_GET_CTX();
  
  switch (mode)
  {
    case DF_ISO_WRAPPING_OFF    : break;
    case DF_ISO_WRAPPING_CARD   : break;

    case DF_ISO_WRAPPING_READER :
    default                     : return DFCARD_LIB_CALL_ERROR;    
  }

  ctx->iso_wrapping = mode;
  return DF_OPERATION_OK;
}

/* DesfireAPI/Command
 *
 * NAME
 *   Command
 *
 * DESCRIPTION
 *   Performs a T=CL block exchange according to ISO 14443-A-4
 *   "low" level command<->response transfert
 *
 * INPUTS
 *   BOOL fast          : if true, DESFire optimized loop is used
 *   WORD accept_len    : length of accepted response
 *   BYTE accept_status : flag to set which status are accepted ACCEPT_OPERATION_OK or ACCEPT_ADDITIONAL_FRAME
 *
 * RETURNS
 *   DF_OPERATION_OK    : T=CL block exchange succeeded
 *   Other code if internal or communication error has occured. 
 *
 **/
SPROX_API_FUNC(Desfire_Command) (SPROX_PARAM  BOOL fast, WORD accept_len, BYTE accept_status)
{
  SPROX_RC   status;
  BYTE    card_status;
  WORD    len = DESFIRE_BUFFER_SIZE;
  SPROX_DESFIRE_GET_CTX();

//  D(printf("S: "));
//  D(dump_buffer(ctx->buffer, ctx->buflen);
//  D(printf("\n"));

  /* Communicate the prepared buffer to the PICC and get its response. */
  if (fast)
  {
    status = SPROX_API_CALL(TclA_ExchangeDF) (SPROX_PARAM_P  ctx->tcl_cid, ctx->buffer, ctx->buflen, ctx->buffer, &len);
  } else
  {
    status = SPROX_API_CALL(TclA_Exchange) (SPROX_PARAM_P  5, ctx->tcl_cid, 0xFF, ctx->buffer, ctx->buflen, ctx->buffer, &len);
  }
  if (status != MI_OK)
  {
    /* Error: could not exchange info blocks with the PICC. */
    return status;
  }

  /* Remember received size */
  ctx->buflen = len;

//  D(printf("R: "));
//  D(dump_buffer(ctx->buffer, ctx->buflen);
//  D(printf("\n"));

  /* Check whether the PICC's response consists of at least one byte.
     If the response is empty, we can not even determine the PICC's status.
     Therefore an empty response is always a length error. */
  if (len < 1)
  {
    /* Error: block with inappropriate number of bytes received from the PICC. */
    return DFCARD_WRONG_LENGTH;
  }

  /*  Check the status byte received from the PICC.
     Depending on which of the flags (sgbACCEPT_OPERATION_OK and sgbACCEPT_ADDITIONAL_FRAME)
     are set in the bExpectedStatusFlags parameter, we accept either the status byte 
     sgbOPERATION_OK or the status byte sgbDF_ADDITIONAL_FRAME or both of them. */
  card_status = ctx->buffer[INF + 0];

  if (!(((card_status == DF_OPERATION_OK) && (accept_status & ACCEPT_OPERATION_OK)) ||
        ((card_status == DF_ADDITIONAL_FRAME) && (accept_status & ACCEPT_ADDITIONAL_FRAME))))
  {
    /* Error: error or unexpected status received from the PICC. */
    return (DFCARD_ERROR - card_status);
  }

  /* Finally check the length of the response. It must be > 0 (already checked) and
     <= sgbMAX_INFO_FRAME_SIZE in all cases.
     If bExpectedResponseLength is != 0, we expect a response with exactly that length.
     If bExpectedResponseLength == 0, we allow any response length > 0 and 
     <= sgbMAX_INFO_FRAME_SIZE. */
  if (((!fast) && (len > MAX_INFO_FRAME_SIZE)) || ((accept_len != 0) && (accept_len != len)))
  {
    /* Error: block with inappropriate number of bytes received from the PICC. */
    return DFCARD_WRONG_LENGTH;
  }

  /* Return the status code from the PICC without any prefix code. */
  return status;
}

#endif