/**h* DesfireAPI/DataManagement
 *
 * NAME
 *   DesfireAPI ::  Data management of DesFire cards
 *
 * COPYRIGHT
 *   (c) 2009 SpringCard - www.springcard.com
 *
 * DESCRIPTION
 *   Get/Set data on DesFire cards
  *   For Mifare DesFire MF3 IC D40 cards
 *
 **/


#include "sprox_desfire_i.h"

/**f* DesfireAPI/ReadData
 *
 * NAME
 *   ReadData
 *
 * DESCRIPTION
 *   Allows to read data from Standard Data File or Backup Data File
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_ReadData(BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     DWORD from_offset,
 *                                     DWORD max_count,
 *                                     BYTE data[],
 *                                     DWORD *done_count);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_ReadData(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     DWORD from_offset,
 *                                     DWORD max_count,
 *                                     BYTE data[],
 *                                     DWORD *done_count);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_ReadData(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     DWORD from_offset,
 *                                     DWORD max_count,
 *                                     BYTE data[],
 *                                     DWORD *done_count);
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   BYTE comm_mode    : file's communication settings (DF_COMM_MODE_PLAIN, DF_COMM_MODE_MACED, 
 *                       DF_COMM_MODE_PLAIN2 or DF_COMM_MODE_ENCIPHERED)(see chapter 3.2 of 
 *                       datasheet of mifare DesFire MF3ICD40 for more information)
 *   DWORD from_offset : starting position for the read operation
 *   DWORD max_count   : maximum data length to read. Set to 0 to read whole file
 *   BYTE data[]       : buffer to receive the data
 *   DWORD *done_count : actual data length read
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, data has been read
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   ReadData2
 *
 **/
SPROX_API_FUNC(Desfire_ReadData) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, DWORD from_offset, DWORD max_count, BYTE data[], DWORD *done_count)
{
  return SPROX_API_CALL(Desfire_ReadDataEx) (SPROX_PARAM_P  DF_READ_DATA, file_id, comm_mode, from_offset, max_count, 1, data, done_count);
}

/**f* DesfireAPI/ReadData2
 *
 * NAME
 *   ReadData2
 *
 * DESCRIPTION
 *   Allows to read data from Standard Data File or Backup Data File
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_ReadData2(BYTE file_id,
 *                                     DWORD from_offset,
 *                                     DWORD max_count,
 *                                     BYTE data[],
 *                                     DWORD *done_count);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_ReadData2(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     DWORD from_offset,
 *                                     DWORD max_count,
 *                                     BYTE data[],
 *                                     DWORD *done_count);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_ReadData2(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     DWORD from_offset,
 *                                     DWORD max_count,
 *                                     BYTE data[],
 *                                     DWORD *done_count);
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   DWORD from_offset : starting position for the read operation in bytes
 *   DWORD max_count   : maximum data length to read. Set to 0 to read whole file
 *   BYTE data[]       : buffer to receive the data
 *   DWORD *done_count : actual data length read
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, data has been read
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   ReadData
 *
 **/
SPROX_API_FUNC(Desfire_ReadData2) (SPROX_PARAM  BYTE file_id, DWORD from_offset, DWORD max_count, BYTE data[], DWORD *done_count)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    read_only_access;
  BYTE    read_write_access;
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* we have to receive the communications mode first */
  status = SPROX_API_CALL(Desfire_GetFileSettings) (SPROX_PARAM_P  file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* Depending on the AccessRights field (settings r and r/w) we have to decide whether     */
  /* we are able to communicate in the mode indicated by comm_mode.                         */
  /* If ctx->auth_key does neither match r nor r/w and one of this settings                 */
  /* contains the value "ever" (0x0E) communication has to be done in plain mode regardless */
  /* of the mode indicated by comm_mode.                                                    */
  read_only_access = (BYTE) ((access_rights & DF_READ_ONLY) >> DF_GET_READ_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((read_only_access != ctx->auth_key) && (read_write_access != ctx->auth_key) && ((read_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return SPROX_API_CALL(Desfire_ReadData) (SPROX_PARAM_P  file_id, comm_mode, from_offset, max_count, data, done_count);
}

/**f* DesfireAPI/WriteData
 *
 * NAME
 *   WriteData
 *
 * DESCRIPTION
 *   Allows to write data from Standard Data File or Backup Data File
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_WriteData(BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     DWORD from_offset,
 *                                     DWORD size,
 *                                     const BYTE data[]);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_WriteData(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     DWORD from_offset,
 *                                     DWORD size,
 *                                     const BYTE data[]);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_WriteData(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     DWORD from_offset,
 *                                     DWORD size,
 *                                     const BYTE data[]);
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   BYTE comm_mode    : file's communication settings (DF_COMM_MODE_PLAIN, DF_COMM_MODE_MACED, 
 *                       DF_COMM_MODE_PLAIN2 or DF_COMM_MODE_ENCIPHERED)(see chapter 3.2 of 
 *                       datasheet of mifare DesFire MF3ICD40 for more information)
 *   DWORD from_offset : starting position for the write operation in bytes
 *   DWORD size        : size of the buffer in bytes
 *   BYTE data[]       : buffer to write to the card
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, data has been written
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   WriteData2
 *
 **/
SPROX_API_FUNC(Desfire_WriteData) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, DWORD from_offset, DWORD size, const BYTE data[])
{
  return SPROX_API_CALL(Desfire_WriteDataEx) (SPROX_PARAM_P  DF_WRITE_DATA, file_id, comm_mode, from_offset, size, data);
}

/**f* DesfireAPI/WriteData2
 *
 * NAME
 *   WriteData2
 *
 * DESCRIPTION
 *   Allows to write data from Standard Data File or Backup Data File
 *
  * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_WriteData2(BYTE file_id,
 *                                     DWORD from_offset,
 *                                     DWORD size,
 *                                     const BYTE data[]);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_WriteData2(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     DWORD from_offset,
 *                                     DWORD size,
 *                                     const BYTE data[]);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_WriteData2(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     DWORD from_offset,
 *                                     DWORD size,
 *                                     const BYTE data[]);
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   DWORD from_offset : starting position for the write operation in bytes
 *   DWORD size        : size of the buffer in bytes
 *   BYTE data[]       : buffer to write to the card
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, data has been written
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   WriteData
 *
 **/
SPROX_API_FUNC(Desfire_WriteData2) (SPROX_PARAM  BYTE file_id, DWORD from_offset, DWORD size, const BYTE data[])
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    write_only_access;
  BYTE    read_write_access;
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* we have to receive the communications mode first */
  status = SPROX_API_CALL(Desfire_GetFileSettings) (SPROX_PARAM_P  file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* Depending on the access_rights field (settings w and r/w) we have to decide whether   */
  /* we are able to communicate in the mode indicated by comm_mode.                        */
  /* If ctx->auth_key does neither match w nor r/w and one of this settings                */
  /* contains the value "ever" (0xE) communication has to be done in plain mode regardless */
  /* of the mode indicated by comm_mode.                                                   */
  write_only_access = (BYTE) ((access_rights & DF_WRITE_ONLY) >> DF_GET_WRITE_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((write_only_access != ctx->auth_key) && (read_write_access != ctx->auth_key) && ((write_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return SPROX_API_CALL(Desfire_WriteData) (SPROX_PARAM_P  file_id, comm_mode, from_offset, size, data);
}

/**f* DesfireAPI/GetValue
 *
 * NAME
 *   GetValue
 *
 * DESCRIPTION
 *   Allows to read current stored value from Value Files
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_GetValue(BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     SDWORD *value);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_GetValue(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     SDWORD *value);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_GetValue(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     SDWORD *value);
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   BYTE comm_mode    : file's communication settings (DF_COMM_MODE_PLAIN, DF_COMM_MODE_MACED, 
 *                       DF_COMM_MODE_PLAIN2 or DF_COMM_MODE_ENCIPHERED)(see chapter 3.2 of 
 *                       datasheet of mifare DesFire MF3ICD40 for more information)
 *   SDWORD *value     : pointer to receive current value
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, value has been read
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   GetValue2
 *
 **/
SPROX_API_FUNC(Desfire_GetValue) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, SDWORD *value)
{
  SDWORD  temp;
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* Create the info block containing the command code and the file ID argument. */
  ctx->buffer[INF + 0] = DF_GET_VALUE;
  ctx->buffer[INF + 1] = file_id;
  ctx->buflen = 2;

  if (comm_mode == DF_COMM_MODE_ENCIPHERED)
  {
    /* The communication mode is DF_COMM_MODE_ENCIPHERED.                   */
    /* GetValue returns 9 bytes, the status byte and a cryptogram block     */
    /* containing the four byte value, 2 CRC bytes and two 0 padding bytes. */
    status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 9, ACCEPT_OPERATION_OK);
    if (status != DF_OPERATION_OK)
      return status;

    /* Decipher the received block. */
    SPROX_API_CALL(Desfire_CipherRecv) (SPROX_PARAM_P  &ctx->buffer[INF + 1], 1, ctx->session_key);

    if (ComputeIso14443ACrc(NULL, &ctx->buffer[INF + 1], 4 + 2))
    {
      /* abortion due to integrity error -> wrong CRC */
      return DFCARD_WRONG_CRC;
    }

    /* Check also the padding bytes for enhanced security. */
    if (ctx->buffer[INF + 7] || ctx->buffer[INF + 8])
    {
      /* Error: cryptogram contains incorrect padding bytes. */
      return DFCARD_WRONG_PADDING;
    }
  } else if (comm_mode == DF_COMM_MODE_MACED)
  {
    BYTE    mac32[4];

    /* The communication mode is DF_COMM_MODE_MACED.                      */
    /* GetValue returns 9 bytes, the status byte, the four byte value and */
    /* the four byte MAC.                                                 */
    status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 9, ACCEPT_OPERATION_OK);
    if (status != DF_OPERATION_OK)
      return status;

    /* Check the received MAC. */
    SPROX_API_CALL(Desfire_CipherMACb) (SPROX_PARAM_P  &ctx->buffer[INF + 1], 4, ctx->session_key, mac32);
    if (memcmp(mac32, &ctx->buffer[INF + 5], 4))
    {
      /* abortion due to integrity error -> wrong MAC */
      return DFCARD_WRONG_MAC;
    }
  } else
  {
    /* GetValue returns 5 bytes, the status byte and the four byte value. */
    status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 5, ACCEPT_OPERATION_OK);
    if (status != DF_OPERATION_OK)
      return status;
  }

  /* Return the requested value bytes. */
  temp = ctx->buffer[INF + 4];
  temp <<= 8;
  temp += ctx->buffer[INF + 3];
  temp <<= 8;
  temp += ctx->buffer[INF + 2];
  temp <<= 8;
  temp += ctx->buffer[INF + 1];
  if (value != NULL)
    *value = temp;

  /* Success. */
  return DF_OPERATION_OK;
}

/**f* DesfireAPI/GetValue2
 *
 * NAME
 *   GetValue2
 *
 * DESCRIPTION
 *   Allows to read current stored value from Value Files
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_GetValue2(BYTE file_id,
 *                                     SDWORD *value);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_GetValue2(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     SDWORD *value);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_GetValue2(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     SDWORD *value);
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   SDWORD *value     : pointer to receive current value
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, value has been read
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   GetValue
 *
 **/
SPROX_API_FUNC(Desfire_GetValue2) (SPROX_PARAM  BYTE file_id, SDWORD *value)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    read_only_access;
  BYTE    write_only_access;
  BYTE    read_write_access;
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* we have to receive the communications mode first */
  status = SPROX_API_CALL(Desfire_GetFileSettings) (SPROX_PARAM_P  file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* GetValue is controlled by the fields r, w, r/w within the access_rights.               */
  /* Depending on the access_rights field (settings r, w and r/w) we have to decide whether */
  /* we are able to communicate in the mode indicated by comm_mode.                         */
  /* If ctx->auth_key does neither match r, w nor r/w and one of this settings              */
  /* contains the value "ever" (0xE) communication has to be done in plain mode regardless  */
  /* of the mode indicated by comm_mode.                                                    */
  read_only_access = (BYTE) ((access_rights & DF_READ_ONLY) >> DF_GET_READ_ONLY);
  write_only_access = (BYTE) ((access_rights & DF_WRITE_ONLY) >> DF_GET_WRITE_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((read_only_access != ctx->auth_key)
      && (write_only_access != ctx->auth_key)
      && (read_write_access != ctx->auth_key) && ((read_only_access == 0x0E) || (write_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return SPROX_API_CALL(Desfire_GetValue) (SPROX_PARAM_P  file_id, comm_mode, value);
}

/**f* DesfireAPI/ReadRecords
 *
 * NAME
 *   ReadRecords
 *
 * DESCRIPTION
 *   Allows to read data out a set of complete records from a Cyclic or Linear Record File.
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_ReadRecords(BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     DWORD from_record,
 *                                     DWORD max_record_count,
 *                                     DWORD record_size,
 *                                     BYTE data[],
 *                                     DWORD *record_count);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_ReadRecords(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     DWORD from_record,
 *                                     DWORD max_record_count,
 *                                     DWORD record_size,
 *                                     BYTE data[],
 *                                     DWORD *record_count);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_ReadRecords(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     DWORD from_record,
 *                                     DWORD max_record_count,
 *                                     DWORD record_size,
 *                                     BYTE data[],
 *                                     DWORD *record_count);
 *
 * INPUTS
 *   BYTE file_id           : File IDentifier
 *   BYTE comm_mode         : file's communication settings (DF_COMM_MODE_PLAIN, DF_COMM_MODE_MACED, 
 *                            DF_COMM_MODE_PLAIN2 or DF_COMM_MODE_ENCIPHERED)(see chapter 3.2 of 
 *                            datasheet of mifare DesFire MF3ICD40 for more information)
 *   DWORD from_record      : offset of the newest record to read. Set to 0 for latest record
 *   DWORD max_record_count : number of records to be read from the PICC. Set to 0 to read all records.
 *   DWORD record_size      : size of the record in bytes
 *   BYTE data[]            : buffer to receive the data
 *   DWORD *record_count    : actual number of records read
 *
 * RETURNS
 *   DF_OPERATION_OK        : success, data has been read
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   ReadRecords2
 *
 **/
SPROX_API_FUNC(Desfire_ReadRecords) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, DWORD from_record, DWORD max_record_count, DWORD record_size, BYTE data[], DWORD *record_count)
{
  BYTE    file_type;
  DWORD   done_size;
  DF_ADDITIONAL_FILE_SETTINGS file_settings;
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  memset(&file_settings, 0x00, sizeof(file_settings));

  /* if a pointer was passed for retrieving the number of records read    */
  /* and no record size was specified, we have to get the record size via */
  /* the GetFileSettings command                                          */
  if ((record_count != NULL) && (record_size == 0))
  {
    status = SPROX_API_CALL(Desfire_GetFileSettings) (SPROX_PARAM_P  file_id, &file_type, 0, 0, &file_settings);
    if (status != DF_OPERATION_OK)
      return status;

    /* if the file type indicates that this is not a record file */
    /* we are not able to proceed                                */
    if ((file_type != DF_LINEAR_RECORD_FILE) && (file_type != DF_CYCLIC_RECORD_FILE))
      return DFCARD_WRONG_FILE_TYPE;

    record_size = file_settings.stRecordFileSettings.eRecordSize;

    /* if eRecordSize becomes zero under any circumstances this     */
    /* would lead to a division by zero at the end of this function */
    /* thus making it impossible to proceed                         */
    if (record_size == 0)
      return DFCARD_WRONG_RECORD_SIZE;
  }

  /* Call the ReadFile function */
  status = SPROX_API_CALL(Desfire_ReadDataEx) (SPROX_PARAM_P  DF_READ_RECORDS, file_id, comm_mode, from_record, max_record_count, record_size, data, &done_size);
  if (status != DF_OPERATION_OK)
    return status;

  /* we have to check whether we received multiples of record_size   */
  /* if this is not the case this means that a format error occurred */
  if ((done_size % record_size) != 0)
    return DFCARD_WRONG_LENGTH;

  /* calculate the number of records if a pointer has been passed */
  if (record_count != NULL)
    *record_count = done_size / record_size;

  return DF_OPERATION_OK;
}

/**f* DesfireAPI/ReadRecords2
 *
 * NAME
 *   ReadRecords2
 *
 * DESCRIPTION
 *   Allows to read data out a set of complete records from a Cyclic or Linear Record File.
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_ReadRecords2(BYTE file_id,
 *                                     DWORD from_record,
 *                                     DWORD max_record_count,
 *                                     DWORD record_size,
 *                                     BYTE data[],
 *                                     DWORD *record_count);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_ReadRecords2(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     DWORD from_record,
 *                                     DWORD max_record_count,
 *                                     DWORD record_size,
 *                                     BYTE data[],
 *                                     DWORD *record_count);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_ReadRecords2(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     DWORD from_record,
 *                                     DWORD max_record_count,
 *                                     DWORD record_size,
 *                                     BYTE data[],
 *                                     DWORD *record_count);
 *
 * INPUTS
 *   BYTE file_id           : File IDentifier
 *   DWORD from_record      : offset of the newest record to read. Set to 0 for latest record
 *   DWORD max_record_count : number of records to be read from the PICC. Set to 0 to read all records.
 *   DWORD record_size      : size of the record in bytes
 *   BYTE data[]            : buffer to receive the data
 *   DWORD *record_count    : actual number of records read
 *
 * RETURNS
 *   DF_OPERATION_OK        : success, data has been read
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   ReadRecords
 *
 **/
SPROX_API_FUNC(Desfire_ReadRecords2) (SPROX_PARAM  BYTE file_id, DWORD from_record, DWORD max_record_count, DWORD record_size, BYTE data[], DWORD *record_count)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    read_only_access;
  BYTE    read_write_access;
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* we have to receive the communications mode first */
  status = SPROX_API_CALL(Desfire_GetFileSettings) (SPROX_PARAM_P  file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* Depending on the AccessRights field (settings r and r/w) we have to decide whether     */
  /* we are able to communicate in the mode indicated by comm_mode.                         */
  /* If ctx->auth_key does neither match r nor r/w and one of this settings                 */
  /* contains the value "ever" (0x0E) communication has to be done in plain mode regardless */
  /* of the mode indicated by comm_mode.                                                    */
  read_only_access = (BYTE) ((access_rights & DF_READ_ONLY) >> DF_GET_READ_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((read_only_access != ctx->auth_key) && (read_write_access != ctx->auth_key) && ((read_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return SPROX_API_CALL(Desfire_ReadRecords) (SPROX_PARAM_P  file_id, comm_mode, from_record, max_record_count, record_size, data, record_count);
}

/**f* DesfireAPI/WriteRecord
 *
 * NAME
 *   WriteRecord
 *
 * DESCRIPTION
 *   Allows to write data to a record in a Cyclic or Linear Record File.
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_WriteRecord(BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     DWORD from_offset,
 *                                     DWORD size,
 *                                     const BYTE data[]);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_WriteRecord(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     DWORD from_offset,
 *                                     DWORD size,
 *                                     const BYTE data[]);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_WriteRecord(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     DWORD from_offset,
 *                                     DWORD size,
 *                                     const BYTE data[]); 
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   BYTE comm_mode    : file's communication settings (DF_COMM_MODE_PLAIN, DF_COMM_MODE_MACED, 
 *                       DF_COMM_MODE_PLAIN2 or DF_COMM_MODE_ENCIPHERED)(see chapter 3.2 of 
 *                       datasheet of mifare DesFire MF3ICD40 for more information)
 *   DWORD from_offset : offset within one single record in bytes
 *   DWORD size        : size data to be written in bytes
 *   BYTE data[]       : buffer containing the data to write
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, data has been written
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   WriteRecord2
 *
 **/
SPROX_API_FUNC(Desfire_WriteRecord) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, DWORD from_offset, DWORD size, const BYTE data[])
{
  return SPROX_API_CALL(Desfire_WriteDataEx) (SPROX_PARAM_P  DF_WRITE_RECORD, file_id, comm_mode, from_offset, size, data);
}

/**f* DesfireAPI/WriteRecord2
 *
 * NAME
 *   WriteRecord2
 *
 * DESCRIPTION
 *   Allows to write data to a record in a Cyclic or Linear Record File.
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_WriteRecord2(BYTE file_id,
 *                                     DWORD from_offset,
 *                                     DWORD size,
 *                                     const BYTE data[]);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_WriteRecord2(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     DWORD from_offset,
 *                                     DWORD size,
 *                                     const BYTE data[]);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_WriteRecord2(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     DWORD from_offset,
 *                                     DWORD size,
 *                                     const BYTE data[]); 
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   DWORD from_offset : offset within one single record in bytes
 *   DWORD size        : size data to be written in bytes
 *   BYTE data[]       : buffer containing the data to write
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, data has been written
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   WriteRecord
 *
 **/
SPROX_API_FUNC(Desfire_WriteRecord2) (SPROX_PARAM  BYTE file_id, DWORD from_offset, DWORD size, const BYTE data[])
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    write_only_access;
  BYTE    read_write_access;
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* we have to receive the communications mode first */
  status = SPROX_API_CALL(Desfire_GetFileSettings) (SPROX_PARAM_P  file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* Depending on the access_rights field (settings w and r/w) we have to decide whether   */
  /* we are able to communicate in the mode indicated by comm_mode.                        */
  /* If ctx->auth_key does neither match w nor r/w and one of this settings                */
  /* contains the value "ever" (0xE) communication has to be done in plain mode regardless */
  /* of the mode indicated by comm_mode.                                                   */
  write_only_access = (BYTE) ((access_rights & DF_WRITE_ONLY) >> DF_GET_WRITE_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((write_only_access != ctx->auth_key) && (read_write_access != ctx->auth_key) && ((write_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return SPROX_API_CALL(Desfire_WriteRecord) (SPROX_PARAM_P  file_id, comm_mode, from_offset, size, data);
}

/**f* DesfireAPI/LimitedCredit
 *
 * NAME
 *   LimitedCredit
 *
 * DESCRIPTION
 *   Allows a limited increase of a value stored in a Value File without having full Read&Write permissions to the file.
 *   This feature can be enabled or disabled during value file creation.
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_LimitedCredit(BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     SDWORD amount);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_LimitedCredit(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     SDWORD amount);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_LimitedCredit(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     SDWORD amount);
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   BYTE comm_mode    : file's communication settings (DF_COMM_MODE_PLAIN, DF_COMM_MODE_MACED, 
 *                       DF_COMM_MODE_PLAIN2 or DF_COMM_MODE_ENCIPHERED)(see chapter 3.2 of 
 *                       datasheet of mifare DesFire MF3ICD40 for more information)
 *   SDWORD amount     : amount to increase to the current value stored in the file. Only positive values allowed.
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, data has been written
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   LimitedCredit2
 *
 **/
SPROX_API_FUNC(Desfire_LimitedCredit) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, SDWORD amount)
{
  return SPROX_API_CALL(Desfire_ModifyValue) (SPROX_PARAM_P  DF_LIMITED_CREDIT, file_id, comm_mode, amount);
}

/**f* DesfireAPI/LimitedCredit2
 *
 * NAME
 *   LimitedCredit2
 *
 * DESCRIPTION
 *   Allows a limited increase of a value stored in a Value File without having full Read&Write permissions to the file.
 *   This feature can be enabled or disabled during value file creation.
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_LimitedCredit2(BYTE file_id,
 *                                     SDWORD amount);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_LimitedCredit2(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     SDWORD amount);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_LimitedCredit2(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     SDWORD amount);
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   SDWORD amount     : amount to increase to the current value stored in the file. Only positive values allowed.
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, data has been written
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   LimitedCredit
 *
 **/
SPROX_API_FUNC(Desfire_LimitedCredit2) (SPROX_PARAM  BYTE file_id, SDWORD amount)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    write_only_access;
  BYTE    read_write_access;
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* we have to receive the communications mode first */
  status = SPROX_API_CALL(Desfire_GetFileSettings) (SPROX_PARAM_P  file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* LimitedCredit is controlled by the fields w and r/w within the access_rights.         */
  /* Depending on the access_rights field (settings w and r/w) we have to decide whether   */
  /* we are able to communicate in the mode indicated by comm_mode.                        */
  /* If ctx->auth_key does neither match w nor r/w and one of this settings                */
  /* contains the value "ever" (0xE) communication has to be done in plain mode regardless */
  /* of the mode indicated by comm_mode.                                                   */
  write_only_access = (BYTE) ((access_rights & DF_WRITE_ONLY) >> DF_GET_WRITE_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((write_only_access != ctx->auth_key) && (read_write_access != ctx->auth_key) && ((write_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return SPROX_API_CALL(Desfire_LimitedCredit) (SPROX_PARAM_P  file_id, comm_mode, amount);
}

/**f* DesfireAPI/Credit
 *
 * NAME
 *   Credit
 *
 * DESCRIPTION
 *   Allows to increase a value stored in a Value File.
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_Credit(BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     SDWORD amount);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_Credit(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     SDWORD amount);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_Credit(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     SDWORD amount);
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   BYTE comm_mode    : file's communication settings (DF_COMM_MODE_PLAIN, DF_COMM_MODE_MACED, 
 *                       DF_COMM_MODE_PLAIN2 or DF_COMM_MODE_ENCIPHERED)(see chapter 3.2 of 
 *                       datasheet of mifare DesFire MF3ICD40 for more information)
 *   SDWORD amount     : amount to increase to the current value stored in the file. Only positive values allowed.
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, data has been written
 *   Other code if internal or communication error has occured. 
 *
 * NOTES
 *   The Credit command requires Authentication with the key specified for "Read&Write" access.
 *
 * SEE ALSO
 *   Credit2
 *
 **/
SPROX_API_FUNC(Desfire_Credit) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, SDWORD amount)
{
  return SPROX_API_CALL(Desfire_ModifyValue) (SPROX_PARAM_P  DF_CREDIT, file_id, comm_mode, amount);
}

/**f* DesfireAPI/Credit2
 *
 * NAME
 *   Credit2
 *
 * DESCRIPTION
 *   Allows to increase a value stored in a Value File.
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_Credit2(BYTE file_id,
 *                                     SDWORD amount);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_Credit2(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     SDWORD amount);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_Credit2(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     SDWORD amount);
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   SDWORD amount     : amount to increase to the current value stored in the file. Only positive values allowed.
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, data has been written
 *   Other code if internal or communication error has occured. 
 *
 * NOTES
 *   The Credit command requires Authentication with the key specified for "Read&Write" access.
 *
 * SEE ALSO
 *   Credit
 *
 **/
SPROX_API_FUNC(Desfire_Credit2) (SPROX_PARAM  BYTE file_id, SDWORD amount)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    read_write_access;
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* we have to receive the communications mode first */
  status = SPROX_API_CALL(Desfire_GetFileSettings) (SPROX_PARAM_P  file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* Credit is controlled by the field r/w within the access_rights.                    */
  /* Depending on the AccessRights field (setting r/w) we have to decide whether        */
  /* we are able to communicate in the mode indicated by comm_mode.                     */
  /* If ctx->auth_key does not match r/w and this setting contains the value            */
  /* "ever" (0xE), communication has to be done in plain mode regardless of the mode    */
  /* indicated by comm_mode.                                                            */
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((read_write_access != ctx->auth_key) && (read_write_access == 0x0E))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return SPROX_API_CALL(Desfire_Credit) (SPROX_PARAM_P file_id, comm_mode, amount);
}

/**f* DesfireAPI/Debit
 *
 * NAME
 *   Debit
 *
 * DESCRIPTION
 *   Allows to decrease a value stored in a Value File.
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_Debit(BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     SDWORD amount);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_Debit(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     SDWORD amount);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_Debit(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     BYTE comm_mode,
 *                                     SDWORD amount);
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   BYTE comm_mode    : file's communication settings (DF_COMM_MODE_PLAIN, DF_COMM_MODE_MACED, 
 *                       DF_COMM_MODE_PLAIN2 or DF_COMM_MODE_ENCIPHERED)(see chapter 3.2 of 
 *                       datasheet of mifare DesFire MF3ICD40 for more information)
 *   SDWORD amount     : amount to decrease to the current value stored in the file. Only positive values allowed.
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, data has been written
 *   Other code if internal or communication error has occured. 
 *
 * NOTES
 *   The Credit command requires Authentication with the key specified for "Read", "Write" ord "Read&Write" access.
 *
 * SEE ALSO
 *   Debit2
 *
 **/
SPROX_API_FUNC(Desfire_Debit) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, SDWORD amount)
{
  return SPROX_API_CALL(Desfire_ModifyValue) (SPROX_PARAM_P  DF_DEBIT, file_id, comm_mode, amount);
}

/**f* DesfireAPI/Debit2
 *
 * NAME
 *   Debit2
 *
 * DESCRIPTION
 *   Allows to decrease a value stored in a Value File.
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_Debit2(BYTE file_id,
 *                                     SDWORD amount);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_Debit2(SPROX_INSTANCE rInst,
 *                                     BYTE file_id,
 *                                     SDWORD amount);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_Debit2(SCARDHANDLE hCard,
 *                                     BYTE file_id,
 *                                     SDWORD amount);
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *   SDWORD amount     : amount to decrease to the current value stored in the file. Only positive values allowed.
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, data has been written
 *   Other code if internal or communication error has occured. 
 *
 * NOTES
 *   The Credit command requires Authentication with the key specified for "Read", "Write" ord "Read&Write" access.
 *
 * SEE ALSO
 *   Debit
 *
 **/
SPROX_API_FUNC(Desfire_Debit2) (SPROX_PARAM  BYTE file_id, SDWORD amount)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    read_only_access;
  BYTE    write_only_access;
  BYTE    read_write_access;
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* we have to receive the communications mode first */
  status = SPROX_API_CALL(Desfire_GetFileSettings) (SPROX_PARAM_P  file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* Debit is controlled by the fields r, w, r/w within the access_rights.                  */
  /* Depending on the access_rights field (settings r, w and r/w) we have to decide whether */
  /* we are able to communicate in the mode indicated by comm_mode.                         */
  /* If ctx->auth_key does neither match r, w nor r/w and one of this settings              */
  /* contains the value "ever" (0xE) communication has to be done in plain mode regardless  */
  /* of the mode indicated by comm_mode.                                                    */
  read_only_access = (BYTE) ((access_rights & DF_READ_ONLY) >> DF_GET_READ_ONLY);
  write_only_access = (BYTE) ((access_rights & DF_WRITE_ONLY) >> DF_GET_WRITE_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((read_only_access != ctx->auth_key)
      && (write_only_access != ctx->auth_key)
      && (read_write_access != ctx->auth_key) && ((read_only_access == 0x0E) || (write_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return SPROX_API_CALL(Desfire_Debit) (SPROX_PARAM_P  file_id, comm_mode, amount);
}

/**f* DesfireAPI/ClearRecordFile
 *
 * NAME
 *   Desfire_ClearRecordFile
 *
 * DESCRIPTION
 *   Allows to reset a Cyclic or Linear Record File to the empty state
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_ClearRecordFile(BYTE file_id);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_ClearRecordFile(SPROX_INSTANCE rInst,
 *                                     BYTE file_id);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_ClearRecordFile(SCARDHANDLE hCard,
 *                                     BYTE file_id);
 *
 * INPUTS
 *   BYTE file_id      : File IDentifier
 *
 * RETURNS
 *   DF_OPERATION_OK   : success, data has been written
 *   Other code if internal or communication error has occured. 
 *
 * NOTES
 *   Full "Read&Write" permission on the file is necessary for executing this command
 *
 **/
SPROX_API_FUNC(Desfire_ClearRecordFile) (SPROX_PARAM  BYTE file_id)
{
  SPROX_DESFIRE_GET_CTX();
  ctx->buffer[INF + 0] = DF_CLEAR_RECORD_FILE;
  ctx->buffer[INF + 1] = file_id;
  ctx->buflen = 2;

  /* response must be of one byte length if successfully terminated */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 1, ACCEPT_OPERATION_OK);
}



